Require Import ssreflect ssrfun ssrbool ssrnat seq eqtype.
Require Import prelude prefix heaps terms mtac xfindM.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


(******************************************************************************)
(* cancelR :                                                                  *)
(*   Lemma automated with Canonical Structures to cancel heap expressions.    *)
(******************************************************************************)


(* The algorithm works as follow:
   - if the heap is h1 :+ h2 then recurse over both and concatenate the results
   - if the heap is the empty heap, return []
   - if the heap is p :-> v then add p to the context, and return [Pts x v], 
     where x is the deBruijn index for p in the context
   - if the heap is whatever else, add the heap to the context and return
     [Var n], where n is the deBruijn index for the heap in the context
 *)

Definition invariant i j t h := [/\ interp j t = h, subctx i j & valid j t].

(* Main structure
   i : input context
   j : output context
   t : syntactification of heap_of using context j *) 
Structure ast (i : ctx) (h : heap) := 
  Ast {synheap_of : synheap; 
       ctx_of : ctx;
       _ : invariant i ctx_of synheap_of h}.

Lemma union_pf i h1 h2 (f1 : ast i h1) (f2 : ast ( (ctx_of f1)) h2) :
        invariant i ( (ctx_of f2)) ( (synheap_of f1) ++ ( (synheap_of f2))) (h1 :+ h2).
Proof.
move: f1 f2=>[t1 j /= [<- S1 D1]] [t2 k /= [<- S2 D2]].
split; first by rewrite interp_cat (interp_subctx D1 S2).
- by apply: (subctx_trans S1 S2).
by rewrite valid_cat D2 andbT; apply: (valid_subctx S2).
Qed.

Lemma empty_pf i : invariant i i [::] empty.
Proof. split; by [|apply: subctx_refl|]. Qed.

Lemma pts_pf A hs xs1 p (d : A) (xs : xfind p xs1):
        invariant (Context hs xs1) (Context hs ( (seq_of xs))) 
                  [:: Pts ( (index_of xs)) (dyn d)] (p :-> d).
Proof.
case: xs=>xs2 x /= [H P]; split; first by rewrite /= H.
- by split; [apply: prefix_refl|].
by apply/andP; rewrite /= (onth_size H). 
Qed.

Lemma var_pf hs1 xs h (f : xfind h hs1) : 
        invariant (Context hs1 xs) (Context ( (seq_of f)) xs) [:: Var ( (index_of f))] h.
Proof.
case:f=>hs2 n [H1 H2]; split; first by rewrite /= /hlook H1.
- by split; [|apply: prefix_refl].
by apply/andP; rewrite /= (onth_size H1).
Qed.

(** Mtactic to reflect a heap into syntax *)
Definition to_ast :=
  mfix f i (h : heap) : M (ast i h) :=
  mmatch h with
  | empty => retS (Ast (empty_pf i) : ast i empty)
  | [x A (v : A)] x :-> v =>
    r <- findx x (ptr_ctx i);
    retS (Ast (i:=i) (pts_pf (heap_ctx i) v r)) 
  | [l r] l :+ r =>
    lp <- f i l;
    rp <- f (ctx_of lp) r;
    retS (Ast (h:=l:+r) (union_pf rp)) 
  | [h] h =>
    r <- findx h (heap_ctx i);
    retS (Ast (i:=i) (var_pf (ptr_ctx i) r)) 
  end. 
Implicit Arguments to_ast [].

Notation cancellation f1 f2
  := (terms.eval (ctx_of f2) (cancel (ctx_of f2) (synheap_of f1) (synheap_of f2))).

Lemma cancelR' h1 h2 
      (f1 : ast empc h1) (f2 : ast ( (ctx_of f1)) h2) :
  def h1 ->
  h1 = h2 ->
  cancellation f1 f2.
Proof.
case: f1 f2 =>hp1 j /= [<- _ I] [hp2 k /= [<- S _]] D H.
by apply: cancel_sound; rewrite -(interp_subctx I S).
Qed.

Set Printing Existential Instances.

(** We need to pack the dependencies with the resulting type *)
Structure packC h1 h2 
  := PackC { 
    f1 : ast empc h1;
    f2 : ast (ctx_of f1) h2;
    cancel : cancellation f1 f2 }.

(** The cancellation tactic *)
Definition cancelR h1 h2 :=
  fun H1 : def h1 =>
  fun H2 : h1 = h2 =>
     (p1 <- to_ast empc h1;
      p2 <- to_ast (ctx_of p1) h2;
      retS (PackC (cancelR' p2 H1 H2) : (packC h1 h2)) 
     ).

(************)
(* Examples *)  
(************)
Example ex0 x (v1 v2:nat): 
          def (x :-> v1) -> x :-> v1 = x :-> v2 -> 
          v1 = v2.
move=>D H.
Time set H' := (cancel (run (cancelR D H))).
Time by rewrite (dyn_inj H').
Time Qed.

Set Printing Implicit.

Example ex1 x h (v1 v2:nat): 
          def (x :-> v1 :+ h) -> x :-> v1 :+ h = x :-> v2 -> 
          if v1 == v2 then true else false.
move=>D H.
Time set H' := run (ret (cancel (run (cancelR D H)))).
Time by rewrite (dyn_inj (proj2 H')) eq_refl.
Time Qed.

Example ex2 h3 h4 x1 (d1 d2 d3 d4 : nat) : 
     def ((h3 :+ (x1 :-> d1)) :+ empty) ->
     (h3 :+ (x1 :-> d1) :+ empty) = 
     h4 :+ (x1 :-> d4) ->
     d1 = d4 /\ h3 = h4.
move=>D H.
Time set (H' := run (retS (cancel (run (cancelR D H))))).
simpl in H'.
Time case: H'=>/= [->].
Time by move/dyn_inj.
Time Qed.


Example ex3 h1 h3 x1 x2 (d1 d2 d3 d4 : nat) : 
     def ((h3 :+ (x1 :-> d1)) :+ h1 :+ (x2 :-> d2)) ->
     (h3 :+ (x1 :-> d1)) :+ h1 :+ (x2 :-> d2) = 
     (x2 :-> d3) :+ h3 :+ (x1 :-> d4) ->
     d1 = d4 /\ d2 = d3 /\ h1 = empty.
move=>D H.
Time set H' := run (retS (cancel (run (cancelR D H)))).
simpl in H'.
Time case: H'=>/=[->] [].
by move/dyn_inj=>->; move/dyn_inj=>->.
Time Qed.

Example ex4 h1 h2 h3 x1 x2 (d1 d2 d3 d4 : nat) : 
     def ((h3 :+ (x1 :-> d1)) :+ (h1 :+ empty) :+ (x2 :-> d2)) ->
     (h3 :+ (x1 :-> d1)) :+ (h1 :+ empty) :+ (x2 :-> d2) = 
     (x2 :-> d3) :+ (h2 :+ h3) :+ (x1 :-> d4) ->
     d1 = d4 /\ d2 = d3 /\ h1 = h2.
move=>D H.
Time set H' := run (retS (cancel (run (cancelR D H)))).
Time case: H'=>/=[->] [].
by move/dyn_inj=>->; move/dyn_inj=>->.
Time Qed.

